/* Copyright (C) Heikki Orsila 2003 <heikki@ee.tut.fi>
   Parses eagleplayer files (in amiga executable format).
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "dpiutil.h"
#include "../osdep/strl.c"

static unsigned int readbig(char *ptr);
static void process_string(char *str);

static unsigned int readbig(char *ptr) {
  unsigned char *p = (unsigned char *) ptr;
  unsigned int x = p[3] + (p[2] << 8) + (p[1] << 16) + (p[0] << 24);
  return x;
}

static void process_string(char *str) {
  int i = 0;
  while (str[i]) {
    if (str[i] == '\n')
      str[i] = ' ';
    i++;
  }
}

int process_eagleplayer(char *credits, char *filename, int credits_len) {
  char buf[1024];
  int i, data_len;
  unsigned int x, y;
  FILE *f;
  unsigned int *tag_table;
  char *hunk;
  int hunk_size;
  int table_size;
  int offset;

  int is_amplifier;
  int is_customplayer;
  int is_noteplayer;
  int has_songend;

  char data[256];

  if (credits)
    *credits = 0;

  if (sizeof(int) != 4) {
    fprintf (stderr, "uade: dpiutil.c: sizeof(int) != 4\n");
    return 0;
  }

  if (!(f = fopen(filename, "rb"))) {
    return 0;
  }
  data_len = fread(buf, 1, sizeof(buf) - 1, f);
  if (data_len <= 0) {
    return 0;
  }
  fclose(f);
  memset(buf + data_len, 0, sizeof(buf) - data_len);

  if (readbig(buf) != 0x000003f3) {
    goto err;
  }

  /* search for hunk data start */
  for (i = 0; i < data_len; i++) {
    /* is data start? */
    if (readbig(buf + i) == 0x70ff4e75) {
      break;
    }
  }

  if (i == data_len) {
    goto err;
  }

  if ((i + 12) >= data_len) {
    goto err;
  }

  if (strncmp(buf + i + 4, "DELIRIUM", 8) != 0 &&
      strncmp(buf + i + 4, "EPPLAYER", 8) != 0    ) {
    goto err;
  }

  if (!credits) {
    printf("filename:\t%s\n", filename);
  }

  hunk = buf + i; /* hunk start */
  hunk_size = data_len - i;

  if ((16 + 5) >= hunk_size) {
    goto oob;
  }

  /* check if $VER is available */
  if (!memcmp(&hunk[16], "$VER:", 5)) {
    offset = 16 + 5;
    while (offset < hunk_size) {
      if (!isspace(hunk[offset]))
	break;
      offset++;
    }
    if (offset >= hunk_size)
      goto oob;
    if ((offset + strlen(hunk + offset) + 1) > ((unsigned int) hunk_size))
      goto oob;
    strlcpy(data, hunk + offset, sizeof(data));
    if (credits) {
      snprintf(credits, credits_len, "VERSION:\n%s\n\n", data);
    } else {
      printf("version:\t%s\n", data);
    }
  }

  is_amplifier = 0;
  is_customplayer = 0;
  is_noteplayer = 0;
  has_songend = 0;

  /* tag table at hunk start + hunk[12] */
  offset = readbig(hunk + 12);
  if (offset < 0) {
    goto err;
  }
  tag_table = (unsigned int *) (hunk + offset);

  if (((char *) tag_table) >= &buf[data_len]) {
    goto oob;
  }

  table_size = ((int) (&buf[data_len] - ((char *) tag_table))) / 8;

  if (table_size <= 0) {
    goto oob;
  }

  /* check all tags in this loop */
  for (i = 0; i < table_size; i += 2) {
    x = readbig((char *) (tag_table + i));
    y = readbig((char *) (tag_table + i + 1));
    
    if (!x)
      break;
    
    switch (x) {
    case 0x80004459:
      if (y >= ((unsigned int) hunk_size))
	goto oob;
      if ((y + strlen(hunk + y) + 1) > ((unsigned int) hunk_size))
	goto oob;
      strlcpy(data, hunk + y, sizeof(data));
      if (credits) {
	strlcat(credits, "PLAYERNAME:\n", credits_len);
	strlcat(credits, data, credits_len);
	strlcat(credits, "\n\n", credits_len);
      } else { 
	printf("playername:\t%s\n", data);
      }
      break;
      
    case 0x8000445a:
      if (y >= ((unsigned int) hunk_size))
	goto oob;
      if ((y + strlen(hunk + y) + 1) > ((unsigned int) hunk_size))
	goto oob;
      strlcpy(data, hunk + y, sizeof(data));
      if (credits) {
	strlcat(credits, "CREDITS:\n", credits_len);
	strlcat(credits, data, credits_len);
	strlcat(credits, "\n\n", credits_len);
      } else {
	process_string(data);
	printf("credits:\t%s\n", data);
      }
      
    case 0x80004474:
      if ((y & 0x2) != 0) {
	has_songend = 1;
      }
      break;
      
    case 0x80004573:
      is_amplifier = 1;
      break;
      
    case 0x80004455:
      is_customplayer = 1;
      break;
      
    case 0x80004479:
    case 0x8000447A:
      is_noteplayer = 1;
      break;
      
    default:
      break;
    }
  }
  
  if (!credits) {
    printf("song end:\t%s\n", has_songend ? "yes" : "no");
    if (is_amplifier)
      printf("amplifier:\tyes\n");
    if (is_customplayer)
      printf("custom player:\tyes\n");
    if (is_noteplayer)
      printf("noteplayer:\tyes\n");
    printf("\n");
  }

  return 1;

 oob:
  fprintf(stderr, "dpi: %s out of boundary\n", filename);
  return 0;

 err:
  fprintf(stderr, "dpi: %s is not an eagleplayer\n", filename);
  return 0;
}
